
import os
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
from scipy.sparse import lil_matrix
from scipy.sparse.linalg import spsolve
# 可视化结果（需要适当的x, y坐标映射）
import plotly.graph_objects as go
import plotly.io as pio
from scipy.spatial import Delaunay
from src.pcl_postproc import smooth_pcl, align2ref
from src.eval import fit_plane, remove_noise
from scipy.ndimage import gaussian_filter
from scipy.optimize import least_squares
import scipy.sparse as sp
from scipy.spatial import Delaunay


def save_scatter_plot(x, y, z):
 
    # 创建三维散点图
    scatter = go.Scatter3d(
        x=x[:, 2],
        y=y[:, 2],
        z=z[:, 2],
        mode='markers',
        marker=dict(
            size=4,              # 点的大小
            color=z[:, 2],     # 使用 z 轴的值作为颜色映射
            colorscale='Viridis',  # 颜色映射
            opacity=0.8          # 点的透明度
        )
    )

    # 创建图表布局
    layout = go.Layout(
        title="Interactive 3D Scatter Plot",
        scene=dict(
            xaxis=dict(title="X-axis"),
            yaxis=dict(title="Y-axis"),
            zaxis=dict(title="Z-axis")
        )
    )

    # 生成图表
    fig = go.Figure(data=[scatter], layout=layout)

    # 保存为HTML文件，可以在浏览器中查看
    pio.write_html(fig, file='interactive_3d_scatter.html', auto_open=True)



# def build_laplacian_matrix_new(points, triangles):
#     N = len(points)
#     A = lil_matrix((N, N))
#     for tri in triangles:
#         for i in range(3):
#             i1, i2, i3 = tri[i], tri[(i+1)%3], tri[(i+2)%3]
#             p1, p2, p3 = points[i1], points[i2], points[i3]
#             cot_alpha = cotangent(p2, p3, p1)
#             cot_beta = cotangent(p1, p3, p2)
#             A[i1, i2] -= cot_alpha
#             A[i1, i3] -= cot_beta
#             A[i2, i1] -= cot_alpha
#             A[i2, i3] -= cot_beta
#             A[i1, i1] += cot_alpha + cot_beta
#     return A.tocsr()
# def cotangent(p1, p2, p3):
#     v1 = p2 - p1
#     v2 = p3 - p1
#     cos_theta = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
#     return cos_theta / np.sqrt(1 - cos_theta**2)


def cotangent(p1, p2, p3, epsilon=1e-6):
    v1 = p2 - p1
    v2 = p3 - p1
    cos_theta = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
    
    # 限制 cos_theta 的范围，防止出现数值不稳定
    cos_theta = np.clip(cos_theta, -1 + epsilon, 1 - epsilon)
    
    # 计算 cotangent
    return cos_theta / np.sqrt(1 - cos_theta**2)


def build_laplacian_matrix_with_boundary(points, triangles, boundary_indices=None):
    N = len(points)
    A = lil_matrix((N, N))
    for tri in triangles:
        for i in range(3):
            i1, i2, i3 = tri[i], tri[(i+1)%3], tri[(i+2)%3]
            p1, p2, p3 = points[i1], points[i2], points[i3]
            cot_alpha = cotangent(p2, p3, p1)
            cot_beta = cotangent(p1, p3, p2)
            A[i1, i2] -= cot_alpha
            A[i1, i3] -= cot_beta
            A[i2, i1] -= cot_alpha
            A[i2, i3] -= cot_beta
            A[i1, i1] += cot_alpha + cot_beta
    
    # Apply boundary conditions
    if boundary_indices is not None:
        for idx in boundary_indices:
            A[idx, :] = 0  # Clear row
            A[idx, idx] = 1  # Fix the boundary value
    
    return A.tocsr()


def build_laplacian_matrix_with_regularization(points, triangles, lambda_reg=1e-4):
    N = len(points)
    A = lil_matrix((N, N))
    for tri in triangles:
        for i in range(3):
            i1, i2, i3 = tri[i], tri[(i+1)%3], tri[(i+2)%3]
            p1, p2, p3 = points[i1], points[i2], points[i3]
            cot_alpha = cotangent(p2, p3, p1)
            cot_beta = cotangent(p1, p3, p2)
            A[i1, i2] -= cot_alpha
            A[i1, i3] -= cot_beta
            A[i2, i1] -= cot_alpha
            A[i2, i3] -= cot_beta
            A[i1, i1] += cot_alpha + cot_beta
    
    # Add regularization to the diagonal
    A.setdiag(A.diagonal() + lambda_reg)
    
    return A.tocsr()


def find_boundary_indices(points, triangles):
    # 创建一个字典来存储边及其对应的出现次数
    edge_count = {}

    # 遍历所有三角形，统计每条边出现的次数
    for tri in triangles:
        for i in range(3):
            # 获取三角形的边，由两个顶点组成
            edge = tuple(sorted([tri[i], tri[(i+1)%3]]))
            if edge in edge_count:
                edge_count[edge] += 1
            else:
                edge_count[edge] = 1
    
    # 边界边的特点是它们只属于一个三角形
    boundary_edges = [edge for edge, count in edge_count.items() if count == 1]

    # 获取边界边的顶点索引
    boundary_indices = set()
    for edge in boundary_edges:
        boundary_indices.update(edge)
    
    return list(boundary_indices)


def build_laplacian_matrix_new(points, triangles, boundary_indices=None):
    N = len(points)
    A = lil_matrix((N, N))
    for tri in triangles:
        for i in range(3):
            i1, i2, i3 = tri[i], tri[(i+1)%3], tri[(i+2)%3]
            p1, p2, p3 = points[i1], points[i2], points[i3]
            cot_alpha = cotangent(p2, p3, p1)
            cot_beta = cotangent(p1, p3, p2)
            A[i1, i2] -= cot_alpha
            A[i1, i3] -= cot_beta
            A[i2, i1] -= cot_alpha
            A[i2, i3] -= cot_beta
            A[i1, i1] += cot_alpha + cot_beta
    
    # Apply boundary conditions if provided
    if boundary_indices is not None:
        for idx in boundary_indices:
            A[idx, :] = 0  # Clear row
            A[idx, idx] = 1  # Set diagonal to 1 to fix the boundary value
    
    return A.tocsr()


def compute_triangle_area(p1, p2, p3):
    # 使用叉积计算三角形面积
    return 0.5 * np.linalg.norm(np.cross(p2 - p1, p3 - p1))

def compute_divergence(points, triangles, gx, gy):
    n = len(points)
    div_g = np.zeros(n)  # 初始化散度

    # 遍历所有三角形
    for tri in triangles:
        i1, i2, i3 = tri  # 三角形的三个顶点索引
        p1, p2, p3 = points[i1], points[i2], points[i3]  # 三角形三个顶点的坐标

        # 计算三角形的面积
        area = compute_triangle_area(p1, p2, p3)

        # 对于每个三角形，我们使用每个顶点的梯度（gx, gy）来计算散度
        g1 = np.array([gx[i1], gy[i1]])
        g2 = np.array([gx[i2], gy[i2]])
        g3 = np.array([gx[i3], gy[i3]])

        # 三角形内部的平均梯度
        mean_gradient = (g1 + g2 + g3) / 3.0

        # 散度 = 平均梯度 dot 面积法向量（此处面积代表的是体积贡献）
        triangle_divergence = np.dot(mean_gradient, np.array([1.0, 1.0])) * area

        # 将该三角形的散度分配给三个顶点
        div_g[i1] += triangle_divergence / 3.0
        div_g[i2] += triangle_divergence / 3.0
        div_g[i3] += triangle_divergence / 3.0

    return div_g

def smooth_gradients(gx, gy, sigma=3):
    # 使用高斯滤波器平滑梯度场
    gx_smooth = gaussian_filter(gx, sigma=sigma)
    gy_smooth = gaussian_filter(gy, sigma=sigma)
    return gx_smooth, gy_smooth

def poisson_recons_with_smoothed_gradients(x, y, gx, gy, debug=False):
    # Smooth the gradient fields
    gx_smooth, gy_smooth = smooth_gradients(gx, gy)

    # Proceed with regular Poisson reconstruction using the smoothed gradients
    return poisson_recons_new(x, y, gx_smooth, gy_smooth, debug)


def poisson_recons_new(x, y, gx, gy, lambda_reg=0.01, debug=False):
    n = len(x)
    points = np.vstack((x, y)).T
    tri = Delaunay(points)
    triangles = tri.simplices

    # Build Laplacian matrix based on Delaunay triangulation
    #L = build_laplacian_matrix_new(points, triangles)
    L = build_laplacian_matrix(points, triangles)
    #L = build_laplacian_matrix_with_regularization(points, triangles)
    #boundary_indices = find_boundary_indices(points, triangles)
    #print('boundary_indices = ', boundary_indices)
    #print('points = ', points)
    #L = build_laplacian_matrix_with_boundary(points, triangles, boundary_indices)

    # 可视化点云和边界点
    # if debug:
    #     plt.figure(figsize=(8, 6))
    #     plt.triplot(points[:, 0], points[:, 1], triangles, color='gray', alpha=0.5)  # 画出三角剖分
    #     plt.scatter(points[:, 0], points[:, 1], color='blue', label="Points")         # 画出所有点
    #     plt.scatter(points[boundary_indices, 0], points[boundary_indices, 1], color='red', label="Boundary Points")  # 标记边界点

    #     # 设置图形参数
    #     plt.legend()
    #     plt.title("Visualization of Points and Boundary Points")
    #     plt.xlabel("X")
    #     plt.ylabel("Y")
    #     plt.show()

    # Add a small regularization term to make the solution smoother
    L_reg = L + lambda_reg * lil_matrix(np.eye(n))

    # Calculate divergence from synthetic gradients
    div_g = compute_divergence(points, triangles, gx, gy)

    # Solve the Poisson equation with regularization
    z_reconstructed = spsolve(L_reg, div_g)

    return z_reconstructed


def build_laplacian_matrix(points, triangles):
    N = len(points)
    A = lil_matrix((N, N))
    for tri in triangles:
        for i in range(3):
            A[tri[i], tri[i]] += 3  # 每个顶点累加对角线上的值
            for j in range(3):
                if i != j:
                    A[tri[i], tri[j]] -= 1  # 对应三角形顶点之间建立连接
    return A.tocsr()

def poisson_recons(x, y, gx, gy, debug=False):

    n = len(x)
    # theta = np.linspace(0, 2 * np.pi, n, endpoint=False)
    points = np.vstack((x, y)).T
    # Delaunay triangulation
    tri = Delaunay(points)
    triangles = tri.simplices

    # Build Laplacian matrix based on Delaunay triangulation
    #L = build_laplacian_matrix_new(points, triangles)
    L = build_laplacian_matrix(points, triangles)

    # # Calculate divergence from synthetic gradients
    # div_g = np.zeros(n)
    # for i in range(n):
    #     div_g[i] = gx[i] + gy[i]  # Simple approximation
    
    # Calculate divergence from synthetic gradients (use improved method)
    gx_smooth, gy_smooth = smooth_gradients(gx, gy, 0.02)
    div_g = compute_divergence(points, triangles, gx_smooth, gy_smooth)

    # Solve the Poisson equation using the sparse matrix
    z_reconstructed = spsolve(L, div_g)

    # Color mapping based on the mean value of the z-coordinates at the vertices of each triangle
    # facecolors = z_reconstructed[triangles].mean(axis=1)
    return z_reconstructed



def reconstruction_cumsum(world_points, camera_points, screen_points, debug=False, smooth=False, align=True, denoise=True):

    x = world_points[:, 0]
    y = world_points[:, 1]
    
    # 计算入射光线向量
    incident_vector = np.array([world_points[:, 0] - camera_points[:, 0],
                                world_points[:, 1] - camera_points[:, 1],
                                world_points[:, 2] - camera_points[:, 2]])

    incident_vector = incident_vector / np.linalg.norm(incident_vector, axis=0)

    # 计算反射光线向量
    reflection_vector = np.array([world_points[:, 0] - screen_points[:, 0],
                                  world_points[:, 1] - screen_points[:, 1],
                                  world_points[:, 2] - screen_points[:, 2]])
    
    reflection_vector = reflection_vector / np.linalg.norm(reflection_vector, axis=0)

    # 计算法向量
    normal_vector = incident_vector + reflection_vector
    normal_vector = normal_vector / np.linalg.norm(normal_vector, axis=0)
 
    normal_vector = np.nan_to_num(normal_vector, nan=0.0)

    #print('normal_vector = ', normal_vector)

    # 计算法向量在X方向的梯度
    gradient_x = normal_vector[0]

    # 计算法向量在Y方向的梯度
    gradient_y = normal_vector[1]
    #gradient_x = np.zeros(gradient_x.shape)
    #gradient_y = np.zeros(gradient_x.shape)

    print('abs mean (gradient_x) = ', np.mean(abs(gradient_x)))
    print('abs mean (gradient_y) = ', np.mean(abs(gradient_y)))

    #gradient_x = gradient_x - np.mean(gradient_x)
    #gradient_y = gradient_y - np.mean(gradient_y)
    # import pdb
    # pdb.set_trace()
    # 计算z值，假设初始z值为0
    z = np.zeros(len(x))
    z[0] = 0  # 初始点的z
    
    z_raw = poisson_recons(x, y, gradient_x, gradient_y)
    #z_raw = poisson_recons_with_smoothed_gradients(x, y, gradient_x, gradient_y)
    #z_raw = z

    #print('gradient_x  = ', x.shape, y.shape, gradient_x.shape, gradient_y.shape)
    gradient_xy = np.column_stack((x, y, gradient_x, gradient_y))
    
    if align:
        #points_aligned, rotation_matrix = align2ref(smoothed_points)
        points_aligned, rotation_matrix = align2ref(np.column_stack((x, y, z_raw)))
        raw_data = np.column_stack((x, y, z_raw))
        
        points_aligned[:,2] = points_aligned[:,2] - np.mean(points_aligned[:, 2])
        #print('-- nanmean after alignment: mean - {}; std - {}'.format(np.nanmean(points_aligned[:, 2]), np.nanstd(points_aligned[:, 2])))
    if smooth:
        #print("-- point cloud is being smoothed.")
        points_smoothed = None
        #points_smoothed = smooth_pcl(points_aligned, 3)
        # print('raw_shape = ', raw_data.shape)
        # print('smoothed_shape = ', points_smoothed.shape)
        # print('raw min max = ', np.min(raw_data[:,0]), np.max(raw_data[:,0]), np.min(raw_data[:,1]), np.max(raw_data[:,1]))
        # print('smoothed min max = ', np.min(points_smoothed[:,0]), np.max(points_smoothed[:,0]), np.min(points_smoothed[:,1]), np.max(points_smoothed[:,1]))
    if denoise:
        points_denoise = None
        # z = points_aligned[:, 2]
        # a, b, c = fit_plane(x, y, z)
        # points_denoise = remove_noise(x, y, z, a, b, c, thresh=0.01)

    if debug:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        # 提取 x, y, z 坐标
        x_vals = x
        y_vals = y
        z_vals = z_raw

        # 绘制3D点云
        ax.scatter(x_vals, y_vals, z_vals, c=z_vals, cmap='viridis', marker='o' ,s=1)

        # 设置轴标签和标题
        ax.set_xlabel('X (mm)')
        ax.set_ylabel('Y (mm)')
        ax.set_zlabel('Z (mm)')
        ax.set_title('raw 3D Point Cloud Visualization')

        plt.show()
 
    return np.column_stack((x, y, z_raw)), gradient_xy




